@@ -63,8 +63,6 @@ const µBlock = (( ) => { // jshint ignore:line
         extensionUpdateForceReload: false,
         filterAuthorMode: false,
         filterOnHeaders: false,
-        ignoreRedirectFilters: false,
-        ignoreScriptInjectFilters: false,
         loggerPopupType: 'popup',
         manualUpdateAssetFetchPeriod: 500,
         popupFontSize: 'unset',
@@ -88,14 +86,16 @@ const µBlock = (( ) => { // jshint ignore:line
         alwaysDetachLogger: true,
         autoUpdate: true,
         cloudStorageEnabled: false,
+        cnameUncloakEnabled: true,
         collapseBlocked: true,
         colorBlindFriendly: false,
         contextMenuEnabled: true,
         dynamicFilteringEnabled: false,
-        externalLists: [],
+        externalLists: '',
         firewallPaneMinimized: true,
         hyperlinkAuditingDisabled: true,
         ignoreGenericCosmeticFilters: vAPI.webextFlavor.soup.has('mobile'),
+        importedLists: [],
         largeMediaSize: 50,
         parseAllABPHideFilters: true,
         popupPanelSections: 0b111,
@@ -164,6 +164,7 @@ const µBlock = (( ) => { // jshint ignore:line
         compiledCosmeticSection: 200,
         compiledScriptletSection: 300,
         compiledHTMLSection: 400,
+        compiledHTTPHeaderSection: 500,
         compiledSentinelSection: 1000,
         compiledBadSubsection: 1,
 
