@@ -72,6 +72,32 @@ const onBlockElementInFrame = function(details, tab) {
 
 /******************************************************************************/
 
+const onSubscribeToList = function(details) {
+    let parsedURL;
+    try {
+        parsedURL = new URL(details.linkUrl);
+    }
+    catch(ex) {
+    }
+    if ( parsedURL instanceof URL === false ) { return; }
+    const url = parsedURL.searchParams.get('location');
+    if ( url === null ) { return; }
+    const title = parsedURL.searchParams.get('title') || '?';
+    const hash = µBlock.selectedFilterLists.indexOf(parsedURL) !== -1
+        ? '#subscribed'
+        : '';
+    vAPI.tabs.open({
+        url:
+            `/asset-viewer.html` +
+            `?url=${encodeURIComponent(url)}` +
+            `&title=${encodeURIComponent(title)}` + 
+            `&subscribe=1${hash}`,
+        select: true,
+    });
+};
+
+/******************************************************************************/
+
 const onTemporarilyAllowLargeMediaElements = function(details, tab) {
     if ( tab === undefined ) { return; }
     let pageStore = µBlock.pageStoreFromTabId(tab.id);
@@ -88,6 +114,12 @@ const onEntryClicked = function(details, tab) {
     if ( details.menuItemId === 'uBlock0-blockElementInFrame' ) {
         return onBlockElementInFrame(details, tab);
     }
+    if ( details.menuItemId === 'uBlock0-blockResource' ) {
+        return onBlockElement(details, tab);
+    }
+    if ( details.menuItemId === 'uBlock0-subscribeToList' ) {
+        return onSubscribeToList(details);
+    }
     if ( details.menuItemId === 'uBlock0-temporarilyAllowLargeMediaElements' ) {
         return onTemporarilyAllowLargeMediaElements(details, tab);
     }
@@ -99,17 +131,28 @@ const menuEntries = {
     blockElement: {
         id: 'uBlock0-blockElement',
         title: vAPI.i18n('pickerContextMenuEntry'),
-        contexts: ['all'],
+        contexts: [ 'all' ],
     },
     blockElementInFrame: {
         id: 'uBlock0-blockElementInFrame',
         title: vAPI.i18n('contextMenuBlockElementInFrame'),
-        contexts: ['frame'],
+        contexts: [ 'frame' ],
+    },
+    blockResource: {
+        id: 'uBlock0-blockResource',
+        title: vAPI.i18n('pickerContextMenuEntry'),
+        contexts: [ 'audio', 'frame', 'image', 'video' ],
+    },
+    subscribeToList: {
+        id: 'uBlock0-subscribeToList',
+        title: vAPI.i18n('contextMenuSubscribeToList'),
+        contexts: [ 'link' ],
+        targetUrlPatterns: [ 'abp:*' ],
     },
     temporarilyAllowLargeMediaElements: {
         id: 'uBlock0-temporarilyAllowLargeMediaElements',
         title: vAPI.i18n('contextMenuTemporarilyAllowLargeMediaElements'),
-        contexts: ['all'],
+        contexts: [ 'all' ],
     }
 };
 
@@ -120,24 +163,35 @@ let currentBits = 0;
 const update = function(tabId = undefined) {
     let newBits = 0;
     if ( µBlock.userSettings.contextMenuEnabled && tabId !== undefined ) {
-        let pageStore = µBlock.pageStoreFromTabId(tabId);
+        const pageStore = µBlock.pageStoreFromTabId(tabId);
         if ( pageStore && pageStore.getNetFilteringSwitch() ) {
-            newBits |= 0x01;
+            if ( pageStore.shouldApplySpecificCosmeticFilters(0) ) {
+                newBits |= 0b0001;
+            } else {
+                newBits |= 0b0010;
+            }
             if ( pageStore.largeMediaCount !== 0 ) {
-                newBits |= 0x02;
+                newBits |= 0b0100;
             }
         }
+        newBits |= 0b1000;
     }
     if ( newBits === currentBits ) { return; }
     currentBits = newBits;
-    let usedEntries = [];
-    if ( newBits & 0x01 ) {
+    const usedEntries = [];
+    if ( newBits & 0b0001 ) {
         usedEntries.push(menuEntries.blockElement);
         usedEntries.push(menuEntries.blockElementInFrame);
     }
-    if ( newBits & 0x02 ) {
+    if ( newBits & 0b0010 ) {
+        usedEntries.push(menuEntries.blockResource);
+    }
+    if ( newBits & 0b0100 ) {
         usedEntries.push(menuEntries.temporarilyAllowLargeMediaElements);
     }
+    if ( newBits & 0b1000 ) {
+        usedEntries.push(menuEntries.subscribeToList);
+    }
     vAPI.contextMenu.setEntries(usedEntries, onEntryClicked);
 };
 
