@@ -540,6 +540,15 @@ FilterContainer.prototype.compileSpecificSelector = function(
 
 /******************************************************************************/
 
+FilterContainer.prototype.compileTemporary = function(parser) {
+    return {
+        session: this.sessionFilterDB,
+        selector: parser.result.compiled,
+    };
+};
+
+/******************************************************************************/
+
 FilterContainer.prototype.fromCompiledContent = function(reader, options) {
     if ( options.skipCosmetic ) {
         this.skipCompiledContent(reader);
@@ -894,7 +903,7 @@ FilterContainer.prototype.retrieveGenericSelectors = function(request) {
         return;
     }
 
-    const out = { injected: '', excepted, };
+    const out = { injectedCSS: '', excepted, };
 
     const injected = [];
     if ( simpleSelectors.size !== 0 ) {
@@ -918,9 +927,9 @@ FilterContainer.prototype.retrieveGenericSelectors = function(request) {
         });
     }
 
-    out.injected = injected.join(',\n');
+    out.injectedCSS = `${injected.join(',\n')}\n{display:none!important;}`;
     vAPI.tabs.insertCSS(request.tabId, {
-        code: out.injected + '\n{display:none!important;}',
+        code: out.injectedCSS,
         frameId: request.frameId,
         matchAboutBlank: true,
         runAt: 'document_start',
@@ -955,9 +964,13 @@ FilterContainer.prototype.retrieveSpecificSelectors = function(
         exceptedFilters: [],
         noDOMSurveying: this.needDOMSurveyor === false,
     };
-    const injectedHideFilters = [];
+    const injectedCSS = [];
 
-    if ( options.noCosmeticFiltering !== true ) {
+    if (
+        options.noSpecificCosmeticFiltering !== true ||
+        options.noGenericCosmeticFiltering !== true
+    ) {
+        const injectedHideFilters = [];
         const specificSet = this.$specificSet;
         const proceduralSet = this.$proceduralSet;
         const exceptionSet = this.$exceptionSet;
@@ -1019,8 +1032,29 @@ FilterContainer.prototype.retrieveSpecificSelectors = function(
         if ( specificSet.size !== 0 ) {
             injectedHideFilters.push(Array.from(specificSet).join(',\n'));
         }
+
+        // Some procedural filters are really declarative cosmetic filters, so
+        // we extract and inject them immediately.
         if ( proceduralSet.size !== 0 ) {
-            out.proceduralFilters = Array.from(proceduralSet);
+            for ( const json of proceduralSet ) {
+                const pfilter = JSON.parse(json);
+                if ( pfilter.tasks === undefined ) {
+                    const { action } = pfilter;
+                    if ( action !== undefined && action[0] === ':style' ) {
+                        injectedCSS.push(`${pfilter.selector}\n{${action[1]}}`);
+                        proceduralSet.delete(json);
+                        continue;
+                    }
+                }
+                if ( pfilter.pseudo !== undefined ) {
+                    injectedHideFilters.push(pfilter.selector);
+                    proceduralSet.delete(json);
+                    continue;
+                }
+            }
+            if ( proceduralSet.size !== 0 ) {
+                out.proceduralFilters = Array.from(proceduralSet);
+            }
         }
 
         // Highly generic cosmetic filters: sent once along with specific ones.
@@ -1063,6 +1097,12 @@ FilterContainer.prototype.retrieveSpecificSelectors = function(
             }
         }
 
+        if ( injectedHideFilters.length !== 0 ) {
+            injectedCSS.push(
+                `${injectedHideFilters.join(',\n')}\n{display:none!important;}`
+            );
+        }
+
         // Important: always clear used registers before leaving.
         specificSet.clear();
         proceduralSet.clear();
@@ -1077,10 +1117,11 @@ FilterContainer.prototype.retrieveSpecificSelectors = function(
         runAt: 'document_start',
     };
 
-    if ( injectedHideFilters.length !== 0 ) {
-        out.injectedHideFilters = injectedHideFilters.join(',\n');
-        details.code = out.injectedHideFilters + '\n{display:none!important;}';
-        if ( options.dontInject !== true ) {
+    // Inject all declarative-based filters as a single stylesheet.
+    if ( injectedCSS.length !== 0 ) {
+        out.injectedCSS = injectedCSS.join('\n\n');
+        details.code = out.injectedCSS;
+        if ( request.tabId !== undefined ) {
             vAPI.tabs.insertCSS(request.tabId, details);
         }
     }
@@ -1091,7 +1132,7 @@ FilterContainer.prototype.retrieveSpecificSelectors = function(
         cacheEntry.retrieve('net', networkFilters);
         if ( networkFilters.length !== 0 ) {
             details.code = networkFilters.join('\n') + '\n{display:none!important;}';
-            if ( options.dontInject !== true ) {
+            if ( request.tabId !== undefined ) {
                 vAPI.tabs.insertCSS(request.tabId, details);
             }
         }
@@ -1123,9 +1164,8 @@ FilterContainer.prototype.benchmark = async function() {
         entity: '',
     };
     const options = {
-        noCosmeticFiltering: false,
+        noSpecificCosmeticFiltering: false,
         noGenericCosmeticFiltering: false,
-        dontInject: true,
     };
     let count = 0;
     const t0 = self.performance.now();
