@@ -83,7 +83,7 @@ const loadDashboardPanel = function(pane, first) {
         uDom('.tabButton.selected').toggleClass('selected', false);
         tabButton.classList.add('selected');
         tabButton.scrollIntoView();
-        uDom.nodeFromId('iframe').setAttribute('src', pane);
+        uDom.nodeFromId('iframe').contentWindow.location.replace(pane);
         if ( pane !== 'no-dashboard.html' ) {
             vAPI.localStorage.setItem('dashboardLastVisitedPane', pane);
         }
