@@ -176,7 +176,11 @@ const mergeStrings = function(urls) {
     merged = merged.replace(/^\*+$/, '')
                    .replace(/\*{2,}/g, '*')
                    .replace(/([^*]{1,3}\*)(?:[^*]{1,3}\*)+/g, '$1');
-    return merged;
+
+    // https://github.com/uBlockOrigin/uBlock-issues/issues/1494
+    let pos = merged.indexOf('/');
+    if ( pos === -1 ) { pos = merged.length; }
+    return merged.slice(0, pos).includes('*') ? urls[0] : merged;
 };
 
 /******************************************************************************/
@@ -374,6 +378,7 @@ const filterTypes = {
 const cosmeticFilterFromElement = function(elem) {
     if ( elem === null ) { return 0; }
     if ( elem.nodeType !== 1 ) { return 0; }
+    if ( noCosmeticFiltering ) { return 0; }
 
     if ( candidateElements.indexOf(elem) === -1 ) {
         candidateElements.push(elem);
@@ -397,7 +402,7 @@ const cosmeticFilterFromElement = function(elem) {
     }
 
     // Tag name
-    const tagName = elem.localName;
+    const tagName = CSS.escape(elem.localName);
 
     // Use attributes if still no selector found.
     // https://github.com/gorhill/uBlock/issues/1901
@@ -705,6 +710,7 @@ const filterToDOMInterface = (( ) => {
     // https://github.com/gorhill/uBlock/issues/2515
     //   Remove trailing pseudo-element when querying.
     const fromCompiledCosmeticFilter = function(raw) {
+        if ( noCosmeticFiltering ) { return; }
         if ( typeof raw !== 'string' ) { return; }
         let elems, style;
         try {
@@ -794,7 +800,7 @@ const filterToDOMInterface = (( ) => {
         if ( permanent === false || reCosmeticAnchor.test(lastFilter) === false ) {
             return apply();
         }
-        if ( vAPI.domFilterer instanceof Object === false ) { return; }
+        if ( noCosmeticFiltering ) { return; }
         const cssSelectors = new Set();
         const proceduralSelectors = new Set();
         for ( const { raw } of lastResultset ) {
@@ -805,9 +811,8 @@ const filterToDOMInterface = (( ) => {
             }
         }
         if ( cssSelectors.size !== 0 ) {
-            vAPI.domFilterer.addCSSRule(
-                Array.from(cssSelectors),
-                vAPI.hideStyle,
+            vAPI.domFilterer.addCSS(
+                `${Array.from(cssSelectors).join('\n')}\n{${vAPI.hideStyle}}`,
                 { mustInject: true }
             );
         }
@@ -883,7 +888,15 @@ const elementFromPoint = (( ) => {
         const magicAttr = `${vAPI.sessionId}-clickblind`;
         pickerRoot.setAttribute(magicAttr, '');
         let elem = document.elementFromPoint(x, y);
-        if ( elem === document.body || elem === document.documentElement ) {
+        if (
+            elem === null || /* to skip following tests */
+            elem === document.body ||
+            elem === document.documentElement || (
+                pickerBootArgs.zap !== true &&
+                noCosmeticFiltering &&
+                resourceURLsFromElement(elem).length === 0
+            )
+        ) {
             elem = null;
         }
         // https://github.com/uBlockOrigin/uBlock-issues/issues/380
@@ -1010,6 +1023,7 @@ const startPicker = function() {
     // Try using mouse position
     if (
         pickerBootArgs.mouse &&
+        vAPI.mouseClick instanceof Object &&
         typeof vAPI.mouseClick.x === 'number' &&
         vAPI.mouseClick.x > 0
     ) {
@@ -1038,14 +1052,28 @@ const startPicker = function() {
     const elems = document.getElementsByTagName(tagName);
     for ( const elem of elems  ) {
         if ( elem === pickerRoot ) { continue; }
-        const src = elem[attr];
-        if ( typeof src !== 'string' ) { continue; }
-        if ( (src !== url) && (src !== '' || url !== 'about:blank') ) {
+        const srcs = resourceURLsFromElement(elem);
+        if (
+            (srcs.length !== 0 && srcs.includes(url) === false) ||
+            (srcs.length === 0 && url !== 'about:blank')
+        ) {
             continue;
         }
-        elem.scrollIntoView({ behavior: 'smooth', block: 'start' });
         filtersFrom(elem);
-        return showDialog({ broad: true });
+        if (
+            netFilterCandidates.length !== 0 ||
+            cosmeticFilterCandidates.length !== 0
+        ) {
+            if ( pickerBootArgs.mouse !== true ) {
+                elem.scrollIntoView({
+                    behavior: 'smooth',
+                    block: 'center',
+                    inline: 'center'
+                });
+            }
+            showDialog({ broad: true });
+        }
+        return;
     }
 
     // A target was specified, but it wasn't found: abort.
@@ -1190,12 +1218,9 @@ const onConnectionMessage = function(msg) {
 }
 
 // The DOM filterer will not be present when cosmetic filtering is disabled.
-if (
-    pickerBootArgs.zap !== true &&
-    vAPI.domFilterer instanceof Object === false
-) {
-    return;
-}
+const noCosmeticFiltering =
+    vAPI.domFilterer instanceof Object === false ||
+    vAPI.noSpecificCosmeticFiltering === true;
 
 // https://github.com/gorhill/uBlock/issues/1529
 //   In addition to inline styles, harden the element picker styles by using
