@@ -69,7 +69,7 @@ const initializeTabs = async function() {
             if ( tab.discarded === true ) { continue; }
             const { id, url } = tab;
             µb.tabContextManager.commit(id, url);
-            µb.bindTabToPageStore(id);
+            µb.bindTabToPageStore(id, 'tabCommitted', tab);
             // https://github.com/chrisaljoudi/uBlock/issues/129
             //   Find out whether content scripts need to be injected
             //   programmatically. This may be necessary for web pages which
@@ -106,25 +106,25 @@ const initializeTabs = async function() {
 const onVersionReady = function(lastVersion) {
     if ( lastVersion === vAPI.app.version ) { return; }
 
-    // Since built-in resources may have changed since last version, we
-    // force a reload of all resources.
-    µb.redirectEngine.invalidateResourcesSelfie();
+    vAPI.storage.set({ version: vAPI.app.version });
 
     const lastVersionInt = vAPI.app.intFromVersion(lastVersion);
     if ( lastVersionInt === 0 ) { return; }
 
+    // Since built-in resources may have changed since last version, we
+    // force a reload of all resources.
+    µb.redirectEngine.invalidateResourcesSelfie();
+
     // https://github.com/LiCybora/NanoDefenderFirefox/issues/196
     //   Toggle on the blocking of CSP reports by default for Firefox.
     if (
-        vAPI.webextFlavor.soup.has('firefox') &&
-        lastVersionInt <= 1031003011
+        lastVersionInt <= 1031003011 &&
+        vAPI.webextFlavor.soup.has('firefox')
     ) {
         µb.sessionSwitches.toggle('no-csp-reports', '*', 1);
         µb.permanentSwitches.toggle('no-csp-reports', '*', 1);
         µb.saveHostnameSwitches();
     }
-
-    vAPI.storage.set({ version: vAPI.app.version });
 };
 
 /******************************************************************************/
@@ -160,12 +160,32 @@ const onNetWhitelistReady = function(netWhitelistRaw, adminExtra) {
 // User settings are in memory
 
 const onUserSettingsReady = function(fetched) {
-    // List of external lists is meant to be an array
-    if ( typeof fetched.externalLists === 'string' ) {
-        fetched.externalLists =
+    // `externalLists` will be deprecated in some future, it is kept around
+    // for forward compatibility purpose, and should reflect the content of
+    // `importedLists`.
+    if ( Array.isArray(fetched.externalLists) ) {
+        fetched.externalLists = fetched.externalLists.join('\n');
+        vAPI.storage.set({ externalLists: fetched.externalLists });
+    }
+    if (
+        fetched.importedLists.length === 0 &&
+        fetched.externalLists !== ''
+    ) {
+        fetched.importedLists =
             fetched.externalLists.trim().split(/[\n\r]+/);
     }
 
+    // https://github.com/uBlockOrigin/uBlock-issues/issues/1513
+    //   Transition nicely.
+    //   TODO: remove when current version of uBO is well past 1.34.0.
+    if ( typeof µb.hiddenSettings.cnameUncloak === false ) {
+        fetched.cnameUncloakEnabled = false;
+        µb.hiddenSettings.cnameUncloak = true;
+        µb.saveHiddenSettings();
+    }
+    µb.hiddenSettingsDefault.cnameUncloak = undefined;
+    µb.hiddenSettings.cnameUncloak = undefined;
+
     fromFetch(µb.userSettings, fetched);
 
     if ( µb.privacySettingsSupported ) {
@@ -175,6 +195,14 @@ const onUserSettingsReady = function(fetched) {
             'webrtcIPAddress': !µb.userSettings.webrtcIPAddressHidden
         });
     }
+
+    // https://github.com/uBlockOrigin/uBlock-issues/issues/1513
+    if (
+        vAPI.net.canUncloakCnames &&
+        µb.userSettings.cnameUncloakEnabled === false
+    ) {
+        vAPI.net.setOptions({ cnameUncloakEnabled: false });
+    }
 };
 
 /******************************************************************************/
@@ -304,7 +332,7 @@ try {
 
     // https://github.com/uBlockOrigin/uBlock-issues/issues/1365
     //   Wait for onCacheSettingsReady() to be fully ready.
-    await Promise.all([
+    const [ , , lastVersion ] = await Promise.all([
         µb.loadSelectedFilterLists().then(( ) => {
             log.info(`List selection ready ${Date.now()-vAPI.T0} ms after launch`);
         }),
@@ -317,6 +345,7 @@ try {
         vAPI.storage.get(createDefaultProps()).then(fetched => {
             log.info(`First fetch ready ${Date.now()-vAPI.T0} ms after launch`);
             onFirstFetchReady(fetched, adminExtra);
+            return fetched.version;
         }),
         µb.loadUserSettings().then(fetched => {
             log.info(`User settings ready ${Date.now()-vAPI.T0} ms after launch`);
@@ -326,6 +355,12 @@ try {
             log.info(`PSL ready ${Date.now()-vAPI.T0} ms after launch`);
         }),
     ]);
+
+    // https://github.com/uBlockOrigin/uBlock-issues/issues/1547
+    if ( lastVersion === '0.0.0.0' && vAPI.webextFlavor.soup.has('chromium') ) {
+        vAPI.app.restart();
+        return;
+    }
 } catch (ex) {
     console.trace(ex);
 }
