@@ -24,9 +24,6 @@
     // https://developers.google.com/analytics/devguides/collection/analyticsjs/
     const noopfn = function() {
     };
-    const noopnullfn = function() {
-        return null;
-    };
     //
     const Tracker = function() {
     };
@@ -38,6 +35,7 @@
     const w = window;
     const gaName = w.GoogleAnalyticsObject || 'ga';
     const gaQueue = w[gaName];
+    // https://github.com/uBlockOrigin/uAssets/pull/4115
     const ga = function() {
         const len = arguments.length;
         if ( len === 0 ) { return; }
@@ -46,6 +44,8 @@
         let a = args[len-1];
         if ( a instanceof Object && a.hitCallback instanceof Function ) {
             fn = a.hitCallback;
+        } else if ( a instanceof Function ) {
+            fn = ( ) => { a(ga.create()); };
         } else {
             const pos = args.indexOf('hitCallback');
             if ( pos !== -1 && args[pos+1] instanceof Function ) {
@@ -61,7 +61,9 @@
     ga.create = function() {
         return new Tracker();
     };
-    ga.getByName = noopnullfn;
+    ga.getByName = function() {
+        return new Tracker();
+    };
     ga.getAll = function() {
         return [];
     };
@@ -71,12 +73,31 @@
     w[gaName] = ga;
     // https://github.com/gorhill/uBlock/issues/3075
     const dl = w.dataLayer;
-    if ( dl instanceof Object && dl.hide instanceof Object && typeof dl.hide.end === 'function' ) {
-        dl.hide.end();
+    if ( dl instanceof Object ) {
+        if ( dl.hide instanceof Object && typeof dl.hide.end === 'function' ) {
+            dl.hide.end();
+        }
+        if ( typeof dl.push === 'function' ) {
+            const doCallback = function(item) {
+                if ( item instanceof Object === false ) { return; }
+                if ( typeof item.eventCallback !== 'function' ) { return; }
+                setTimeout(item.eventCallback, 1);
+            };
+            if ( Array.isArray(dl) ) {
+                dl.push = item => doCallback(item);
+                const q = dl.slice();
+                dl.length = 0;
+                for ( const item of q ) {
+                    doCallback(item);
+                }
+            }
+        }
     }
     // empty ga queue
     if ( gaQueue instanceof Function && Array.isArray(gaQueue.q) ) {
-        for ( const entry of gaQueue.q ) {
+        const q = gaQueue.q.slice();
+        gaQueue.q.length = 0;
+        for ( const entry of q ) {
             ga(...entry);
         }
     }
